// Copyright 2022 Paul Greenberg greenpau@outlook.com
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package authn

import (
	"context"
	"encoding/json"
	"fmt"
	"github.com/greenpau/go-authcrunch/pkg/authn/enums/operator"
	"github.com/greenpau/go-authcrunch/pkg/identity"
	"github.com/greenpau/go-authcrunch/pkg/ids"
	"github.com/greenpau/go-authcrunch/pkg/requests"
	"github.com/greenpau/go-authcrunch/pkg/user"
	"net/http"
	"strings"
)

func (p *Portal) handleHTTPSSHKeysSettings(
	ctx context.Context, r *http.Request, rr *requests.Request,
	usr *user.User, store ids.IdentityStore, data map[string]interface{},
) error {
	var action string
	var status bool
	entrypoint := "sshkeys"
	data["view"] = entrypoint
	endpoint, err := getEndpoint(r.URL.Path, "/"+entrypoint)
	if err != nil {
		return err
	}
	switch {
	case strings.HasPrefix(endpoint, "/add") && r.Method == "POST":
		// Add SSH key.
		action = "add"
		status = true
		if err := validateKeyInputForm(r, rr); err != nil {
			attachFailStatus(data, "Bad Request")
			break
		}
		rr.Key.Usage = "ssh"
		if err = store.Request(operator.AddKeySSH, rr); err != nil {
			attachFailStatus(data, fmt.Sprintf("%v", err))
			break
		}
		attachSuccessStatus(data, "Public SSH key has been added")
	case strings.HasPrefix(endpoint, "/add"):
		action = "add"
	case strings.HasPrefix(endpoint, "/delete"):
		// Delete a particular SSH key.
		action = "delete"
		status = true
		keyID, err := getEndpointKeyID(endpoint, "/delete/")
		if err != nil {
			attachFailStatus(data, fmt.Sprintf("%v", err))
			break
		}
		rr.Key.ID = keyID
		if err = store.Request(operator.DeletePublicKey, rr); err != nil {
			attachFailStatus(data, fmt.Sprintf("failed deleting key id %s: %v", keyID, err))
			break
		}
		attachSuccessStatus(data, fmt.Sprintf("key id %s deleted successfully", keyID))
	case strings.HasPrefix(endpoint, "/view"):
		// Get a particular SSH key.
		action = "view"
		keyID, err := getEndpointKeyID(endpoint, "/view/")
		if err != nil {
			attachFailStatus(data, fmt.Sprintf("%v", err))
			break
		}
		rr.Key.Usage = "ssh"
		if err = store.Request(operator.GetPublicKeys, rr); err != nil {
			attachFailStatus(data, fmt.Sprintf("failed fetching key id %s: %v", keyID, err))
			break
		}
		bundle := rr.Response.Payload.(*identity.PublicKeyBundle)
		for _, k := range bundle.Get() {
			if k.ID != keyID {
				continue
			}
			var keyMap map[string]interface{}
			keyBytes, _ := json.Marshal(k)
			json.Unmarshal(keyBytes, &keyMap)
			for _, w := range []string{"payload", "openssh"} {
				if _, exists := keyMap[w]; !exists {
					continue
				}
				delete(keyMap, w)
			}
			prettyKey, _ := json.MarshalIndent(keyMap, "", "  ")
			attachSuccessStatus(data, "OK")
			data["key"] = string(prettyKey)
			if k.Payload != "" {
				data["pem_key"] = k.Payload
			}
			if k.OpenSSH != "" {
				data["openssh_key"] = k.OpenSSH
			}
			break
		}
	default:
		// List SSH Keys.
		rr.Key.Usage = "ssh"
		if err = store.Request(operator.GetPublicKeys, rr); err != nil {
			attachFailStatus(data, fmt.Sprintf("%v", err))
			break
		}
		bundle := rr.Response.Payload.(*identity.PublicKeyBundle)
		pubKeys := bundle.Get()
		if len(pubKeys) > 0 {
			data[entrypoint] = pubKeys
		}
	}
	attachView(data, entrypoint, action, status)
	return nil
}
